from diffgram.brain.inference import Inference

import tempfile

try:
    import cv2
except:
    print("Warning: Could not import cv2. Some SDK functions may not be available.")

try:
    import tensorflow as tf
except:
    print("Warning: Could not import tensorflow. Some SDK functions may not be available")

import numpy as np
import requests
import scipy.misc

import diffgram.utils.visualization_utils as vis_util


class Brain():

    def __init__(
            self,
            client,
            name = None,
            id = None,
            local = False,
            use_temp_storage = True
    ):
        """
        client, project client object
        name, string, exact match for Project AI name
        local, bool, run model locally

        if local is true will perform additional setup work local_setup()

        """

        self.client = client

        if self.client.project_string_id is None:
            raise Exception("\n No project string id in client.")

        self.name = name
        self.id = id
        self.status = None
        self.local = local
        self.method = None
        self.sub_method = None
        self.min_score_thresh = .5
        self.build_complete = None
        self.model_path = None
        self.image_to_run = None

        self.use_temp_storage = use_temp_storage
        self.local_model_storage_path = None

        if self.local is True:
            # These are only needed for local operations
            self.temp = tempfile.mkdtemp()

            self.local_setup()

    def inference_from_response(
            self,
            dict):

        # Assumes object detection
        # TODO condition on method

        inference = Inference(
            method = "object_detection",
            id = dict['id'],
            status = dict['status'],
            box_list = dict['box_list'],
            score_list = dict['score_list'],
            label_list = dict['label_list']
        )

        return inference

    def predict_from_url(
            self,
            url):
        """
        url, string, web end point to get file
        """

        if self.local is True:
            raise Exception("Not supported for local models yet.")

        request = {}
        request['url'] = url
        request['ai_name'] = self.name

        endpoint = "/api/walrus/v1/project/" + self.client.project_string_id + \
                   "/inference/from_url"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request)

        self.client.handle_errors(response)

        data = response.json()

        self.client.handle_errors(response)

        inference = self.inference_from_response(data['inference'])
        return inference

    def predict_from_local(
            self,
            path):
        """
        Make a prediction from a local file.
        Creates a Diffgram file object and runs prediction.

        This is roughly equal to running file.from_local() and predict()
        but in one request (instead of two).

        path, string, file path
        """
        if self.local is True:
            self.image_to_run = open(path, "rb")

            # WIP
            # TODO clean up, declare options for different types of expected inputs
            # this is for model that expects numpy array as input
            # self.image_np = scipy.misc.imread(path)
            # self.image_np = self.resize(self.image_np)

            # moved this here, was part of other thing prior
            self.image_to_run = self.image_to_run.read()

            self.run()

            inference = self.inference_from_local()

            return inference

        if self.local is False:
            files = {'file': open(path, 'rb')}

            options = {'immediate_mode': 'True',
                       'ai_name': self.name}

            endpoint = "/api/walrus/v1/project/" + self.client.project_string_id \
                       + "/inference/from_local"

            response = self.client.session.post(
                self.client.host + endpoint,
                files = files,
                data = options)

            self.client.handle_errors(response)

            data = response.json()

            inference = self.inference_from_response(data['inference'])

            return inference

        # TODO handle creation of Inference and Instance objects

    def run(
            self,
            image = None):

        if self.build_complete is False:
            return False

        if image:
            self.image_to_run = image

        with self.graph.as_default():

            # MUST HAVE compat.as_bytes for tf slim
            # https://www.tensorflow.org/api_docs/python/tf/compat/as_bytes
            # https://stackoverflow.com/questions/46687348/decoding-tfrecord-with-tfslim

            self.image_to_run_expanded = tf.compat.as_bytes(self.image_to_run)

            self.image_to_run_expanded = np.expand_dims(self.image_to_run_expanded, axis = 0)

            self.method = "object_detection"

            if self.sub_method == "default" or self.sub_method is None:
                self.run_object_detection()

        inference = self.inference_from_local()

        return inference

    def run_object_detection(self):

        (boxes, scores, classes, num) = self.sess.run(
            [self.detection_boxes,
             self.detection_scores,
             self.detection_classes,
             self.num_detections],
            feed_dict = {
                self.image_tensor: self.image_to_run_expanded})

        self.boxes = np.squeeze(boxes)
        self.scores = np.squeeze(scores)
        self.classes = np.squeeze(classes).astype(np.int32)

    # print(self.boxes, self.scores, self.classes)

    def nearest_iou(self, alpha, bravo):

        _best_iou_hyper = .2

        for i in range(len(alpha.box_list)):

            best_iou = 0
            best_index = None

            # Find best IoU
            for j in range(len(bravo.box_list)):

                iou = Brain.calc_iou(alpha.box_list[i], bravo.box_list[j])

                if iou >= best_iou:
                    best_iou = iou
                    best_index = j

            if best_index is None:
                continue

            # handle large boxes, is the threat entirely inside the box?
            alpha_box = alpha.box_list[i]

            bravo_box = bravo.box_list[best_index]

            if best_iou > _best_iou_hyper or best_iou > .01 and \
                    alpha_box[1] < bravo_box[1] and \
                    alpha_box[3] > bravo_box[3] and \
                    alpha_box[0] < bravo_box[0] and \
                    alpha_box[2] > bravo_box[2]:
                # Assumes boxes have been thresholded already,
                # This way threshold applies to nearest search too

                class_id = bravo.label_list[best_index]

                nearest_alpha_box = bravo.box_list[best_index]

                # for stats
                # self.average_iou = ( (best_iou + self.average_iou  ) / 2)

                # Where best_index is which bravo one
                # is "in" which i index

                print("alpha is in bravo", i, "in", best_index)

    @staticmethod
    def calc_iou(box_a, box_b):
        # Calculate intersection, i.e. area of overlap between the 2 boxes (could be 0)
        # http://math.stackexchange.com/a/99576
        x_overlap = max(0, min(box_a[2], box_b[2]) - max(box_a[0], box_b[0]))
        y_overlap = max(0, min(box_a[3], box_b[3]) - max(box_a[1], box_b[1]))
        intersection = x_overlap * y_overlap

        # Calculate union
        area_box_a = (box_a[2] - box_a[0]) * (box_a[3] - box_a[1])
        area_box_b = (box_b[2] - box_b[0]) * (box_b[3] - box_b[1])
        union = area_box_a + area_box_b - intersection

        if union == 0:
            return 0

        iou = intersection / union
        return iou

    def resize(self, image):

        if image.shape[0] > 600 or image.shape[1] > 600:
            ratio = min((300 / image.shape[0]),
                        (300 / image.shape[1]))

            shape_x = int(round(image.shape[0] * ratio))
            shape_y = int(round(image.shape[1] * ratio))

            image = scipy.misc.imresize(image,
                                        (shape_x, shape_y))

        # print(image.shape)

        return image

    def predict_from_file(
            self,
            file_id):
        """
        file_id, int, diffgram file id

        Assumes singular file for now
        """

        if self.local is True:
            raise Exception("Not supported for local models yet.")

        request = {}
        request['file_list'] = [{'id': file_id}]
        request['ai_name'] = self.name
        request['wait_for_inference'] = True

        endpoint = "/api/walrus/project/" + self.client.project_string_id + \
                   "/inference/add"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request)

        self.client.handle_errors(response)

        data = response.json()

        inference = self.inference_from_response(data['inference'])
        return inference

    def local_setup(self):
        """
        Intial setup for local prediction
        """

        self.get_checkpoint_and_label_map()
        self.build()

    def get_checkpoint_and_label_map(self):
        """

        Get download links
        Download checkpoint file for AI name

        """
        request = {}
        request['ai_name'] = self.name

        endpoint = "/api/walrus/project/" + self.client.project_string_id + \
                   "/brain/local_info"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request)

        self.client.handle_errors(response)

        data = response.json()
        ai = data['ai']
        self.id = ai['id']

        # TODO continue to try and clarify label map crazinesss

        self.file_id_to_model_id = ai['label_dict']
        # print("Label map", self.file_id_to_model_id)

        self.model_id_to_file_id = {v: k for k, v in self.file_id_to_model_id.items()}
        self.file_id_to_name = {v: k for k, v in self.client.name_to_file_id.items()}

        self.build_model_id_to_name()

        # Data has url for models and label map
        # TODO clarify difference between local path and url to download model

        if self.use_temp_storage is True:
            self.model_path = self.temp + "/" + str(self.id) + ".pb"

        if self.use_temp_storage is False:
            self.model_path = self.local_model_storage_path

        self.url_model = ai['url_model']

        self.download_file(
            url = self.url_model,
            path = self.model_path)

    def build_model_id_to_name(self):
        """Creates dictionary of COCO compatible categories keyed by category  id.
        Args:
        categories: a list of dicts, each of which has the following keys:
        'id': (required) an integer id uniquely identifying this category.
        'name': (required) string representing category name
        e.g., 'cat', 'dog', 'pizza'.
        Returns:
        category_index: a dict containing the same entries as categories, but  keyed
        by the 'id' field of each category.
        """

        self.model_id_to_name = {}

        for file_id, label_name in self.file_id_to_name.items():

            model_id = self.file_id_to_model_id.get(str(file_id), None)

            if model_id:
                self.model_id_to_name[model_id] = {'name': label_name}

    # print(self.model_id_to_name)

    def download_file(
            self,
            url,
            path
    ):

        retry = 0
        while retry < 3:

            if url[0: 4] != "http":
                return False

            response = requests.get(url, stream = True)

            if response.status_code != 200:
                retry += 1

            content_type = response.headers.get('content-type', None)

            with open(path, 'wb') as file:

                file.write(response.content)

            return True

        return False

    def check_status(
            self):
        """


        """

        request = {}
        request['ai_name'] = self.name

        endpoint = "/api/walrus/v1/project/" + self.client.project_string_id + \
                   "/brain/status"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request)

        self.client.handle_errors(response)

        data = response.json()

        self.status = data['ai']['status']

    def clean(self):

        try:
            shutil.rmtree(self.temp)  # delete directory
        except OSError as exc:
            if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
                raise  # re-raise exception

    def build(self):
        """
        Build graph for local prediction

        Assumes it has the checkpoint ready to go

        """

        self.graph = tf.Graph()

        with self.graph.as_default():
            # with tf.device('/cpu:0'): # for local cpu testing
            graph_def = tf.GraphDef()

            with tf.gfile.GFile(self.model_path, 'rb') as fid:
                serialized_graph = fid.read()
                graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(graph_def, name = '')

            self.sess = tf.Session(graph = self.graph)

        # TODO make this more flexible to work with different tensor types
        self.image_tensor = self.graph.get_tensor_by_name('encoded_image_string_tensor:0')
        self.detection_boxes = self.graph.get_tensor_by_name('detection_boxes:0')
        self.detection_scores = self.graph.get_tensor_by_name('detection_scores:0')
        self.detection_classes = self.graph.get_tensor_by_name('detection_classes:0')
        self.num_detections = self.graph.get_tensor_by_name('num_detections:0')

        self.build_complete = True

        return True

    def inference_from_local(
            self, ):

        box_list = []
        score_list = []
        label_list = []

        for i in range(self.boxes.shape[0]):
            if self.scores[i] is None:
                pass
            if self.scores[i] > self.min_score_thresh:
                # print("Detection")

                box_list.append(self.boxes[i].tolist())
                label_list.append(self.classes[i].tolist())
                score_list.append(self.scores[i].tolist())

        inference = Inference(
            method = self.method,
            id = None,
            status = None,
            box_list = box_list,
            score_list = score_list,
            label_list = label_list
        )

        return inference

    def visual(self,
               image = None
               ):

        if image is None:
            image = self.image_backup

        # WIP

        # if self.sub_method == "default" or self.sub_method is None:

        # print("ran visual")

        vis_util.visualize_boxes_and_labels_on_image_array(
            image,
            self.boxes,
            self.classes,
            self.scores,
            self.model_id_to_name,
            use_normalized_coordinates = True,
            line_thickness = 3,
            min_score_thresh = self.min_score_thresh)

        return image



from diffgram.core.directory import Directory


class Role:
    id: int
    name: int
    project_id: int
    permissions_list: list

    def __init__(self,
                 client,
                 role_dict = None):

        self.client = client

        if self.client.project_string_id is None:
            raise Exception("\n No project string id in client.")

        self.refresh_from_dict(
            role_dict = role_dict)

    def refresh_from_dict(self, role_dict: dict = None):

        if not role_dict:
            return

        for key, value in role_dict.items():
            setattr(self, key, value)

    def __repr__(self):
        return f'Role<{str(self.serialize())}>'

    def serialize(self):

        return {
            'id': self.id,
            'name': self.name,
            'project_id': self.project_id,
            'permissions_list': self.permissions_list,
        }

    def new(self, name: str = None) -> 'Role':
        """
            Creates a new Role
        :param name:
        :return:
        """
        endpoint = "/api/v1/project/{}/roles/new".format(self.client.project_string_id)

        response = self.client.session.post(
            self.client.host + endpoint,
            json = {'name': name})

        self.client.handle_errors(response)

        data = response.json()
        role = Role(client = self.client, role_dict = data)
        return role

    def delete(self) -> 'Role':
        """
            Creates a new Role
        :param name:
        :return:
        """
        if not self.id:
            raise Exception('Role has no ID. Cannot be deleted.')

        endpoint = f"/api/v1/project/{self.client.project_string_id}/roles/{self.id}/delete"

        response = self.client.session.delete(
            self.client.host + endpoint, json = {})

        self.client.handle_errors(response)

        data = response.json()
        role = Role(client = self.client, role_dict = data)
        return role

    def add_permission(self, perm: str, object_type: str):
        endpoint = f"/api/v1/project/{self.client.project_string_id}/roles/{self.id}/add-perm"

        response = self.client.session.patch(
            self.client.host + endpoint, json = {
                'permission': perm,
                'object_type': object_type,
            })

        self.client.handle_errors(response)

        data = response.json()
        role = Role(client = self.client, role_dict = data)
        self.permissions_list = role.permissions_list
        return role

    def remove_permission(self, perm: str, object_type: str):
        endpoint = f"/api/v1/project/{self.client.project_string_id}/roles/{self.id}/remove-perm"

        response = self.client.session.patch(
            self.client.host + endpoint, json = {
                'permission': perm,
                'object_type': object_type,
            })

        self.client.handle_errors(response)

        data = response.json()
        role = Role(client = self.client, role_dict = data)
        self.permissions_list = role.permissions_list
        return role

    def assign_to_member_in_object(self, member_id: int, object_id: int, object_type: str):
        endpoint = f"/api/v1/project/{self.client.project_string_id}/role-member-object"

        if not self.id:
            raise Exception(f'Provide role ID')

        response = self.client.session.patch(
            self.client.host + endpoint, json = {
                'member_id': member_id,
                'role_id': self.id,
                'object_type': object_type,
                'object_id': object_id,
            }
        )

        self.client.handle_errors(response)
        role_member_object = response.json()
        return role_member_object

    def remove_role_assignment(self, member_id: int, object_id: int, object_type: str):
        endpoint = f"/api/v1/project/{self.client.project_string_id}/role-member-object/remove"

        if not self.id:
            raise Exception(f'Provide role ID')

        response = self.client.session.patch(
            self.client.host + endpoint, json = {
                'member_id': member_id,
                'role_id': self.id,
                'object_type': object_type,
                'object_id': object_id,
            }
        )

        self.client.handle_errors(response)
        role_member_object = response.json()
        return role_member_object

    def get(self, name: str) -> list:

        endpoint = f"/api/v1/project/{self.client.project_string_id}/roles"

        response = self.client.session.get(self.client.host + endpoint)

        self.client.handle_errors(response)
        roles = response.json()

        result = None
        for r in roles:
            if r.get('name') == name:
                result = r
                break
        if result is None:
            raise Exception(f'Role {name} not found')
        role_obj = Role(client = self.client, role_dict = result)
        return role_obj

    def list(self) -> list:

        endpoint = f"/api/v1/project/{self.client.project_string_id}/roles"
        response = self.client.session.get(self.client.host + endpoint)

        self.client.handle_errors(response)
        roles = response.json()
        result = []
        for r in roles:
            role_obj = Role(client = self.client, role_dict = r)
            result.append(role_obj)
        return result



import logging

from PIL import Image, ImageDraw
from imageio import imread
import numpy as np
import traceback
import sys
from threading import Thread
from concurrent.futures import ThreadPoolExecutor
from typing import Callable


class DiffgramDatasetIterator:
    diffgram_file_id_list: list
    max_size_cache: int = 1073741824
    pool: ThreadPoolExecutor
    project: 'Project'
    file_cache: dict
    _internal_file_list: list
    current_file_index: int
    custom_signer_fn: Callable

    def __init__(self,
                 project,
                 diffgram_file_id_list,
                 validate_ids = True,
                 max_size_cache = 1073741824,
                 max_num_concurrent_fetches = 25,
                 custom_signer_fn = None):
        """

        :param project (sdk.core.core.Project): A Project object from the Diffgram SDK
        :param diffgram_file_list (list): An arbitrary number of file ID's from Diffgram.
        """
        self.diffgram_file_id_list = []
        self.max_size_cache = 1073741824
        self.pool = None
        self.custom_signer_fn = custom_signer_fn
        self.file_cache = {}
        self._internal_file_list = []
        self.current_file_index = 0
        self.start_iterator(
            project = project,
            diffgram_file_id_list = diffgram_file_id_list,
            validate_ids = validate_ids,
            max_size_cache = max_size_cache,
            max_num_concurrent_fetches = max_num_concurrent_fetches
        )

    def start_iterator(self,
                       project,
                       diffgram_file_id_list,
                       validate_ids = True,
                       max_size_cache = 1073741824,
                       max_num_concurrent_fetches = 25):
        self.diffgram_file_id_list = diffgram_file_id_list
        self.max_size_cache = max_size_cache
        self.pool = ThreadPoolExecutor(max_num_concurrent_fetches)
        self.project = project
        self.file_cache = {}
        self._internal_file_list = []
        if validate_ids:
            self.__validate_file_ids()
        self.current_file_index = 0

    def __iter__(self):
        self.current_file_index = 0
        return self

    def __len__(self):
        return len(self.diffgram_file_id_list)

    def save_file_in_cache(self, idx, instance_data):
        # If size of cache greater than 1GB (Default)
        if sys.getsizeof(self.file_cache) > self.max_size_cache:
            keys = list(self.file_cache.keys())
            latest_keys = keys[:-10]  # Get oldest 10 elements
            for k in latest_keys:
                self.file_cache.pop(k)

        self.file_cache[idx] = instance_data

    def get_next_n_items(self, idx, num_items = 25):
        """
            Get next N items and save them to cache proactively.
        :param idx:
        :param n:
        :return:
        """
        latest_index = idx + num_items
        if latest_index >= len(self.diffgram_file_id_list):
            latest_index = len(self.diffgram_file_id_list)

        for i in range(idx + 1, latest_index):
            self.pool.submit(self.__get_file_data_for_index, (i,))
        return True

    def __get_file_data_for_index(self, idx):
        diffgram_file = self.project.file.get_by_id(self.diffgram_file_id_list[idx], with_instances = True,
                                                    use_session = False)
        instance_data = self.get_file_instances(diffgram_file)
        self.save_file_in_cache(idx, instance_data)
        return instance_data

    def __getitem__(self, idx):
        if self.file_cache.get(idx):
            return self.file_cache.get(idx)

        result = self.__get_file_data_for_index(idx)

        self.get_next_n_items(idx, num_items = 25)

        return result

    def __next__(self):
        if self.file_cache.get(self.current_file_index):
            result = self.file_cache.get(self.current_file_index)
            self.current_file_index += 1
            return result
        if self.current_file_index >= len(self.diffgram_file_id_list):
            raise StopIteration
        instance_data = self.__get_file_data_for_index(self.current_file_index)
        self.current_file_index += 1
        return instance_data

    def __validate_file_ids(self):
        if not self.diffgram_file_id_list:
            return
        result = self.project.file.file_list_exists(
            self.diffgram_file_id_list,
            use_session = False)
        if not result:
            raise Exception(
                'Some file IDs do not belong to the project. Please provide only files from the same project.')

    def set_custom_url_signer(self, signer_fn: Callable):
        self.custom_signer_fn = signer_fn

    def get_image_data(self, diffgram_file):
        MAX_RETRIES = 10
        image = None
        if hasattr(diffgram_file, 'image'):
            for i in range(0, MAX_RETRIES):
                try:
                    url = None
                    if diffgram_file.image:
                        url = diffgram_file.image.get('url_signed')
                    if diffgram_file.image and self.custom_signer_fn is not None:
                        blob_path = diffgram_file.image['url_signed_blob_path']
                        bucket_name = diffgram_file.image['bucket_name']
                        url = self.custom_signer_fn(blob_path, bucket_name)
                    if url:
                        image = imread(url)
                    break
                except Exception as e:
                    if i < MAX_RETRIES - 1:
                        continue
                    else:
                        print('Fetch Image Failed: Diffgram File ID: {}'.format(diffgram_file.id))
                        print(traceback.format_exc())
                        return None
            return image
        else:
            raise Exception('Pytorch datasets only support images. Please provide only file_ids from images')

    def gen_global_attrs(self, instance_list):
        res = []
        for inst in instance_list:
            if inst['type'] != 'global':
                continue
            res.append(inst['attribute_groups'])
        return res

    def gen_tag_instances(self, instance_list):
        result = []
        for inst in instance_list:
            if inst['type'] != 'tag':
                continue
            for k in list(inst.keys()):
                val = inst[k]
                if val is None:
                    inst.pop(k)
            elm = {
                'label': inst['label_file']['label']['name'],
                'label_file_id': inst['label_file']['id'],
            }
            result.append(elm)
        return result

    def get_file_instances(self, diffgram_file) -> dict:
        if not diffgram_file:
            return
        sample = {'diffgram_file': diffgram_file, 'type': diffgram_file.type}
        if diffgram_file.type not in ['image', 'frame', 'compound']:
            logging.warning('File type "{}" is not supported yet'.format(diffgram_file.type))
            return sample
        if diffgram_file.type in ['image', 'frame']:
            sample['image'] = self.get_image_data(diffgram_file)
        elif diffgram_file.type  is not None and diffgram_file.type.startswith('compound'):
            from diffgram.file.compound_file import CompoundFile
            compound_file: CompoundFile = diffgram_file
            sample['children'] = []
            child_files = compound_file.fetch_child_files(with_instances = True)
            print('chsad', child_files)
            for child in child_files:
                result = self.get_file_instances(child)
                sample['children'].append(result)
        instance_list = diffgram_file.instance_list
        instance_types_in_file = set([x['type'] for x in instance_list])
        # Process the instances of each file

        has_boxes = False
        has_poly = False
        has_tags = False
        has_global = False
        sample['raw_instance_list'] = instance_list
        if 'box' in instance_types_in_file:
            has_boxes = True
            x_min_list, x_max_list, y_min_list, y_max_list = self.extract_bbox_values(instance_list, diffgram_file)
            sample['x_min_list'] = x_min_list
            sample['x_max_list'] = x_max_list
            sample['y_min_list'] = y_min_list
            sample['y_max_list'] = y_max_list
        else:
            sample['x_min_list'] = []
            sample['x_max_list'] = []
            sample['y_min_list'] = []
            sample['y_max_list'] = []

        if 'polygon' in instance_types_in_file:
            has_poly = True
            mask_list = self.extract_masks_from_polygon(instance_list, diffgram_file)
            sample['polygon_mask_list'] = mask_list
        if 'tag' in instance_types_in_file:
            has_tags = True
            sample['tags'] = self.gen_tag_instances(instance_list)
        if 'global' in instance_types_in_file:
            has_global = True
            sample['global_attributes'] = self.gen_global_attrs(instance_list)

        else:
            sample['polygon_mask_list'] = []

        if len(instance_types_in_file) > 4 and has_poly and has_boxes and has_tags and has_global:
            raise NotImplementedError(
                'SDK Streaming only supports boxes and polygon, tags and global attributes types currently. If you want a new instance type to be supported please contact us!'
            )

        label_id_list, label_name_list = self.extract_labels(instance_list)
        sample['label_id_list'] = label_id_list
        sample['instance_types_in_file'] = instance_types_in_file
        sample['label_name_list'] = label_name_list

        return sample

    def extract_masks_from_polygon(self, instance_list, diffgram_file, empty_value = 0):
        nx, ny = diffgram_file.image['width'], diffgram_file.image['height']
        mask_list = []
        if nx is None or ny is None:
            return mask_list

        for instance in instance_list:
            if instance['type'] != 'polygon':
                continue
            poly = [(p['x'], p['y']) for p in instance['points']]
            img = Image.new(mode = 'L', size = (nx, ny), color = 0)  # mode L = 8-bit pixels, black and white
            draw = ImageDraw.Draw(img)
            draw.polygon(poly, outline = 1, fill = 1)
            mask = np.array(img).astype('float32')
            # mask[np.where(mask == 0)] = empty_value
            mask_list.append(mask)
        return mask_list

    def extract_labels(self, instance_list, allowed_instance_types = None):
        label_file_id_list = []
        label_names_list = []
        for inst in instance_list:
            if inst['type'] == 'global':
                continue
            if inst is None:
                continue
            if allowed_instance_types and inst['type'] in allowed_instance_types:
                continue
            label_file_id_list.append(inst['label_file']['id'])
            label_names_list.append(inst['label_file']['label']['name'])

        return label_file_id_list, label_names_list

    def extract_bbox_values(self, instance_list, diffgram_file):
        """
            Creates a pytorch tensor based on the instance type.
            For now we are assuming shapes here, but we can extend it
            to accept custom shapes specified by the user.
        :param instance:
        :return:
        """
        x_min_list = []
        x_max_list = []
        y_min_list = []
        y_max_list = []

        for inst in instance_list:
            if inst['type'] != 'box':
                continue
            x_min_list.append(inst['x_min'])
            x_max_list.append(inst['x_max'])
            y_min_list.append(inst['y_min'])
            y_max_list.append(inst['y_max'])

        return x_min_list, x_max_list, y_min_list, y_max_list



from diffgram.core.core import Project
from typing import List
from diffgram.file.file import File
from diffgram.job.job import Job
from uuid import uuid4


class CompoundChildFile:
    id: int
    path: str
    url: str
    blob_path: str
    file_name: str
    bucket_name: str
    child_file_type: str
    connection_id: int
    media_type: str
    job: Job
    job_id: int
    directory_id: int
    instance_list: list
    frame_packet_map: dict
    assume_new_instances_machine_made: bool
    convert_names_to_label_files: bool
    video_split_duration: int
    local_id: str
    ordinal: int
    root_file: 'CompoundFile'

    def __init__(self,
                 root_file: 'CompoundFile',
                 child_file_type: str,
                 path: str = None,
                 url: str = None,
                 blob_path: str = None,
                 file_name: str = None,
                 bucket_name: str = None,
                 connection_id: int = None,
                 media_type: str = None,
                 job: Job = None,
                 job_id: int = None,
                 directory_id: int = None,
                 instance_list: list = None,
                 frame_packet_map: dict = None,
                 assume_new_instances_machine_made: bool = None,
                 convert_names_to_label_files: bool = None,
                 video_split_duration: int = None,
                 ordinal: int = 0,
                 id: int = None):
        self.root_file = root_file
        self.child_file_type = child_file_type
        self.path = path
        self.url = url
        self.blob_path = blob_path
        self.file_name = file_name
        self.bucket_name = bucket_name
        self.connection_id = connection_id
        self.media_type = media_type
        self.job = job
        self.job_id = job_id
        self.directory_id = directory_id
        self.instance_list = instance_list
        self.frame_packet_map = frame_packet_map
        self.assume_new_instances_machine_made = assume_new_instances_machine_made
        self.convert_names_to_label_files = convert_names_to_label_files
        self.video_split_duration = video_split_duration
        self.local_id = str(uuid4())
        self.ordinal = ordinal
        self.id = None

    def __str__(self):
        return f'<CompoundChildFile id={self.id} ordinal={self.ordinal} file_name={self.file_name} child_file_type={self.child_file_type}>'
    def refresh_from_data_dict(self, data: dict):
        if not data:
            return
        for key in data:
            setattr(self, key, data[key])

    def update(self):
        """
            Syncs child file data with backend API
        :return:
        """
        payload = {
            'ordinal': self.ordinal,
        }
        client = self.root_file.project
        endpoint = f"/api/v1/project/{self.root_file.project.project_string_id}/file/{self.id}/update-metadata"

        response = client.session.put(
            client.host + endpoint,
            json = payload)

        client.handle_errors(response)

        data = response.json()
        new_file_data = data['file']

        self.refresh_from_data_dict(data = new_file_data)
        return self

    def set_ordinal(self, value: int):
        self.ordinal = value


class CompoundFile:
    project: Project
    parent_file_data: dict
    child_files: List[CompoundChildFile]
    instance_list: List[dict]

    def __init__(self, project: Project, name: str, directory_id: int, instance_list: List[dict] = [], file_type: str = 'compound'):
        self.project = project
        self.name = name
        self.directory_id = directory_id
        self.child_files = []
        self.instance_list = instance_list
        self.type = file_type

    @staticmethod
    def from_dict(project: Project, dir_id: int, dict_data: dict):
        result = CompoundFile(project = project, name = dict_data.get('original_filename'), directory_id = dir_id)
        result.__refresh_compound_file_from_data_dict(data = dict_data)
        child_files = result.fetch_child_files()
        result.child_files = child_files
        return result

    def fetch_child_files(self, with_instances: bool = False) -> List[CompoundChildFile]:
        client = self.project
        endpoint = f"/api/v1/project/{self.project.project_string_id}/file/{self.id}/child-files"

        response = client.session.get(client.host + endpoint, params = {'with_instances': with_instances})

        client.handle_errors(response)

        data = response.json()
        child_files_data = data['child_files']
        print('child_files_data', child_files_data)
        result = []
        for elm in child_files_data:
            child_file = CompoundChildFile(root_file = self, child_file_type = elm.get('type'))
            child_file.refresh_from_data_dict(data = elm)
            result.append(child_file)
        return result

    def update_all(self) -> bool:
        """
            Syncs parent and child metadata with backend API.
        :return: True/False
        """
        for child in self.child_files:
            child.update()
        return True

    def __refresh_compound_file_from_data_dict(self, data: dict):
        if not data:
            return
        for key in data:
            setattr(self, key, data[key])

    def remove_child_file(self, child_file: CompoundChildFile) -> List[CompoundChildFile]:
        self.child_files.remove(child_file)
        return self.child_files

    def __create_compound_parent_file(self):
        url = f'/api/v1/project/{self.project.project_string_id}/file/new-compound'
        data = {
            'name': self.name,
            'directory_id': self.directory_id,
            'instance_list': self.instance_list,
            'type': self.type
        }
        response = self.project.session.post(url = self.project.host + url,
                                             json = data)
        self.project.handle_errors(response)
        data = response.json()
        self.parent_file_data = data.get('file')
        self.__refresh_compound_file_from_data_dict(data.get('file'))
        return data.get('file')

    def __create_child_file(self, child_file: CompoundChildFile):
        if child_file.child_file_type == 'from_local':
            return self.project.file.from_local(
                path = child_file.path,
                directory_id = self.directory_id,
                instance_list = child_file.instance_list,
                frame_packet_map = child_file.frame_packet_map,
                assume_new_instances_machine_made = child_file.assume_new_instances_machine_made,
                convert_names_to_label_files = child_file.convert_names_to_label_files,
                parent_file_id = self.parent_file_data.get('id'),
                ordinal = child_file.ordinal,
            )
        elif child_file.child_file_type == 'from_url':
            return self.project.file.from_url(
                url = child_file.path,
                media_type = child_file.media_type,
                job_id = child_file.job_id,
                job = child_file.job,
                video_split_duration = child_file.video_split_duration,
                instance_list = child_file.instance_list,
                frame_packet_map = child_file.frame_packet_map,
                parent_file_id = self.parent_file_data.get('id'),
                ordinal = child_file.ordinal,
            )
        elif child_file.child_file_type == 'from_blob_path':
            return self.project.file.from_blob_path(
                blob_path = child_file.blob_path,
                bucket_name = child_file.bucket_name,
                connection_id = child_file.connection_id,
                media_type = child_file.media_type,
                instance_list = child_file.instance_list,
                file_name = child_file.file_name,
                frame_packet_map = child_file.frame_packet_map,
                parent_file_id = self.parent_file_data.get('id'),
                ordinal = child_file.ordinal,
            )

    def add_child_from_local(self,
                             path: str,
                             instance_list: list = None,
                             frame_packet_map: dict = None,
                             assume_new_instances_machine_made: bool = True,
                             convert_names_to_label_files: bool = True,
                             ordinal: int = None):
        if ordinal is None:
            ordinal = len(self.child_files)
        new_child_file = CompoundChildFile(
            root_file = self,
            child_file_type = "from_local",
            path = path,
            directory_id = self.directory_id,
            instance_list = instance_list,
            frame_packet_map = frame_packet_map,
            assume_new_instances_machine_made = assume_new_instances_machine_made,
            convert_names_to_label_files = convert_names_to_label_files,
            ordinal = ordinal
        )
        self.child_files.append(new_child_file)
        return new_child_file

    def add_child_file_from_url(self,
                                url: str,
                                media_type: str = "image",
                                job: Job = None,
                                job_id: int = None,
                                video_split_duration: int = None,
                                instance_list: list = None,
                                frame_packet_map: dict = None,
                                ordinal: int = None):
        if ordinal is None:
            ordinal = len(self.child_files)
        new_child_file = CompoundChildFile(
            root_file = self,
            child_file_type = "from_url",
            url = url,
            media_type = media_type,
            job = job,
            directory_id = self.directory_id,
            job_id = job_id,
            video_split_duration = video_split_duration,
            instance_list = instance_list,
            frame_packet_map = frame_packet_map,
            ordinal = ordinal
        )
        self.child_files.append(new_child_file)
        return new_child_file

    def add_child_from_blob_path(self,
                                 blob_path: str,
                                 bucket_name: str,
                                 connection_id: int,
                                 media_type: str = 'image',
                                 instance_list: list = None,
                                 file_name: str = None,
                                 frame_packet_map: dict = None,
                                 ordinal: int = None
                                 ):
        if ordinal is None:
            ordinal = len(self.child_files)
        new_child_file = CompoundChildFile(
            root_file = self,
            child_file_type = "from_blob_path",
            blob_path = blob_path,
            bucket_name = bucket_name,
            connection_id = connection_id,
            directory_id = self.directory_id,
            media_type = media_type,
            instance_list = instance_list,
            file_name = file_name,
            frame_packet_map = frame_packet_map,
            ordinal = ordinal
        )
        self.child_files.append(new_child_file)
        return new_child_file

    def upload(self):
        if len(self.child_files) == 0:
            raise AssertionError('Need to add at least one child file to the compound file before calling upload()')
        parent_file_data: dict = self.__create_compound_parent_file()

        for child_file in self.child_files:
            self.__create_child_file(child_file)
        return parent_file_data



from diffgram.core.directory import Directory


class Job():
    valid_output_dir_actions = ['copy', 'move']
    def __init__(self,
                 client,
                 job_dict = None):

        self.client = client

        if self.client.project_string_id is None:
            raise Exception("\n No project string id in client.")

        # TODO review way to set all properties
        # from existing job update

        self.id = None
        self.name = None
        self.status = None
        self.stat_count_tasks = None
        self.stat_count_complete = None
        self.percent_completed = None
        self.tasks_remaining = None
        self.instance_type = None
        self.share = None
        self.type = None
        self.permission = None
        self.field = None
        self.category = None
        self.review_by_human_freqeuncy = None
        self.label_mode = None
        self.passes_per_file = None
        self.attached_directories = []
        self.member_list_ids = []
        self.attached_directories_dict = None

        self.refresh_from_dict(
            job_dict = job_dict)

    def refresh_from_dict(
            self,
            job_dict = None):

        if not job_dict:
            return

        for key, value in job_dict.items():
            setattr(self, key, value)

    def __repr__(self):
        return str(self.serialize())

    def __add_directory_to_job(self, directory: Directory, mode = 'sync'):
        """

        :param directories: Array of directories Objects
        :param mode: sync, copy_once or inactive.
        :return:
        """
        self.attached_directories.append(
            {
                'directory_id': directory.id,
                'selected': mode,

            }
        )
        return self.attached_directories

    def attach_directories(self, directories, mode = 'sync', override_existing = False):
        """
            Attaches directories to a job.
        :param directories: Array of directories Objects
        :param mode: sync, copy_once or inactive.
        :param override_existing: if True. detaches all existing directories.
        :return: The updated job object with the new directories.
        """
        if len(directories) == 0:
            raise ValueError('Provide at least 1 directory')
        if self.status != 'draft':
            raise ValueError('Can only attach directories while on draft mode, not when job has been already launched.')
        if override_existing:
            self.attached_directories = []

        for dir in directories:
            self.attached_directories.append(
                {
                    'directory_id': dir['id'],
                    'selected': mode

                }
            )
        self.attached_directories_dict = {'attached_directories_list': self.attached_directories}
        endpoint = "/api/v1/project/{}/job/update".format(self.client.project_string_id)
        data = self.serialize()
        data['job_id'] = self.id
        response = self.client.session.post(
            self.client.host + endpoint,
            json = data)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            # TODO review better way to update fields
            self.id = data["job"]["id"]

        return self

    def serialize(self):

        if hasattr(self.launch_datetime, 'isoformat'):
            self.launch_datetime = self.launch_datetime.isoformat()

        label_file_list = None
        if hasattr(self, "label_file_list") and self.label_file_list:
            label_file_list = [file.serialize() for file in self.label_file_list]

        return {
            'id': self.id,
            'name': self.name,
            'status': self.status,
            'label_schema_id': self.label_schema_id,
            'instance_type': self.instance_type,
            'share': self.share,
            'type': self.type,
            'permission': self.permission,
            'field': self.field,
            'category': self.category,
            'review_by_human_freqeuncy': self.review_by_human_freqeuncy,
            'label_mode': self.label_mode,
            'passes_per_file': self.passes_per_file,
            'attached_directories_dict': self.attached_directories_dict,
            'launch_datetime': self.launch_datetime,
            'label_file_list': label_file_list,
            'member_list_ids': self.member_list_ids,
            'tag_list': self.tag_list
        }
    def attach_output_dir(self, dir: Directory, action = 'copy'):

        if action not in self.valid_output_dir_actions:
            raise ValueError(f'Invalid actions. Can only be {self.valid_output_dir_actions}')

        data = {
            'job_id': self.id,
            'output_dir': str(dir.id),
            'output_dir_action': action
        }
        endpoint = "/api/v1/project/{}/job/set-output-dir".format(self.client.project_string_id)
        response = self.client.session.post(
            self.client.host + endpoint,
            json = data)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            # TODO review better way to update fields
            self.id = data["job"]["id"]

        return self


    def new(self,
            name = None,
            instance_type = None,
            share = "project",
            job_type = "Normal",
            label_schema_id = None,
            permission = None,
            field = None,
            category = None,
            review_by_human_freqeuncy = None,
            label_mode = None,
            passes_per_file = None,
            guide = None,
            launch_datetime = None,
            sync_directories = [],
            single_copy_directories = [],
            members_list_ids = [],
            auto_launch = True,
            tag_list = [],
            output_dir = None,
            output_dir_action = None
            ):
        """

        Arguments
            self,
            config, a dict of job data
            launch, bool, Launch job after creation

        Expects

        Returns

        """

        # QUESTION create job object eariler instead of after response?
        if len(members_list_ids) == 0:
            raise ValueError('Please provide at least one member_id in members_list_ids.')
        job = Job(client = self.client)

        if label_schema_id is None:
            if self.client.label_schema_list:
                label_schema_id = self.client.label_schema_list[0].get('id')

        job.name = name
        job.instance_type = instance_type
        job.share = share
        job.type = job_type  # careful rename to type from job_type
        job.permission = permission
        job.field = field
        job.category = category
        job.review_by_human_freqeuncy = review_by_human_freqeuncy
        job.label_mode = label_mode
        job.passes_per_file = passes_per_file
        job.launch_datetime = launch_datetime
        job.label_schema_id = label_schema_id
        job.attached_directories = []
        job.member_list_ids = members_list_ids
        job.tag_list = tag_list

        if len(sync_directories) == 0 and len(single_copy_directories) == 0:
            raise ValueError(
                'Please provide at least one attached directory to the job in either sync_directories param or single_copy_directories')

        for dir in sync_directories:
            job.__add_directory_to_job(directory = dir, mode = 'sync')

        for dir in single_copy_directories:
            job.__add_directory_to_job(directory = dir, mode = 'sync')

        job.attached_directories_dict = {
            'attached_directories_list': job.attached_directories
        }

        endpoint = "/api/v1/project/{}/job/new".format(self.client.project_string_id)

        response = self.client.session.post(
            self.client.host + endpoint,
            json = job.serialize())

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            # TODO review better way to update fields
            job.id = data["job"]["id"]

        if guide:
            job.guide_update(guide = guide)

        if output_dir and output_dir_action:
            if output_dir_action not in self.valid_output_dir_actions:
                raise ValueError(f'Invalid actions. Can only be {self.valid_output_dir_actions}')
            job.attach_output_dir(dir = output_dir, action = output_dir_action)
        if auto_launch:
            endpoint_launch = "/api/v1/job/launch".format(self.client.project_string_id)
            response = self.client.session.post(
                self.client.host + endpoint_launch,
                json = {
                    'job_id': data['job']['id']
                })
            self.client.handle_errors(response)

        return job

    def list(self,
             limit = 100,
             status = ["All"],
             name = None,
             tags = [],
             page_number = 1,
            ):
        """

        :param limit:
        :param status: a list with any of the values: ["All", "draft", "active", "complete", "cancelled"]
        :param tags:
        :return:
        """

        # Example usage print(project.job.list().json())
        tag_id_list = None
        if len(tags) > 0:
            # fetch tag id list
            tag_path = f'/api/v1/project/{self.client.project_string_id}/tags/list'
            response_tags = self.client.session.get(self.client.host + tag_path)
            tag_list = response_tags.json()['tag_list']
            tag_id_list = []
            for elm in tag_list:
                if elm['name'] in tags:
                    tag_id_list.append(elm['id'])
        endpoint = "/api/v1/job/list"
        request_json_body = {"metadata":
            {
                "limit": limit,
                "data_mode": "with_tags",
                "search": name,
                "status": status,
                "tag_list": tag_id_list,
                "project_string_id": self.client.project_string_id,
                "page_number": page_number
            }
        }

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request_json_body)

        self.client.handle_errors(response)
        result = response.json()
        return result['Job_list']

    def launch(
            self
    ):
        """

        Arguments
            self,

        Expects
            None

        Returns
            True if success

        """

        endpoint = "/api/v1/job/launch"

        request = {'job_id': self.id}

        response = self.client.session.post(
            self.client.host + endpoint,
            json = request)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            print("Launched")
            return True

        return False

    def guide_update(
            self,
            guide,
            kind = "default",
            action = "update"
    ):
        """

        Arguments
            self,
            guide, class Guide object
            kind options ["default", "review"]
            update_or_remove options ["update", "remove"]

        Expects

        Returns
            None, prints update

        """

        endpoint = "/api/v1/guide/attach/job"

        update_dict = {'guide_id': guide.id,
                       'job_id': self.id,
                       'kind': kind,
                       'update_or_remove': action}

        response = self.client.session.post(self.client.host + endpoint,
                                            json = update_dict)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            print("Guide update success")
            return True

        return False

    def archive_jobs(self, id_list: list):
        job_list = [{'id': elm} for elm in id_list]
        endpoint = "/api/v1/job/cancel"

        payload = {'job_list': job_list, 'mode': 'archive'}

        response = self.client.session.post(self.client.host + endpoint, json = payload)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            return True

        return False

    def get_by_id(
            self,
            id: int):
        """
        """

        job = Job(client = self.client)
        job.id = id

        job.refresh_info()

        return job

    def generate_export(
            self,
            kind = 'Annotations',
            return_type = "data",
            source = "job",
            masks = False,
            directory_id = None,
            wait_for_export_generation = True,
            ann_is_complete = None  # Bool. None=='all', True=='complete' tasks only
    ):
        """

        Arguments
            self,
            kind, string, in ["Annotations", "TF Records"]
            return_type, string, in ["url", "data"]

        # Note that the "data" return type is for kind "Annotations"
        # The data is expected to be returned in JSON format

        Expects

        Returns

        """

        endpoint = "/api/walrus/project/" + self.client.project_string_id + \
                   "/export/to_file"

        # TODO not a fan of "return_type" variable name
        # Also, can we map this into a more "automatic"
        # Default? ie tf records being a url etc..

        spec_dict = {
            'job_id': self.id,
            'kind': kind,
            'source': source,
            'file_comparison_mode': "latest",
            'directory_id': directory_id,
            'masks': masks,
            'ann_is_complete': ann_is_complete,
            'return_type': return_type,
            'wait_for_export_generation': wait_for_export_generation
        }

        response = self.client.session.post(self.client.host + endpoint,
                                            json = spec_dict)

        self.client.handle_errors(response)

        data = response.json()

        if wait_for_export_generation is False:
            export = self.client.export.new(data.get('export'))
            return export

        return data

    def refresh_info(
            self,
            mode_data = None
    ):
        """
        Assumptions

        Arguments

        Returns

        """

        endpoint = "/api/v1/job/" + str(self.id) + \
                   "/builder/info"

        spec_dict = {
            'mode_data': mode_data
        }

        response = self.client.session.post(
            self.client.host + endpoint,
            json = spec_dict)

        self.client.handle_errors(response)

        data = response.json()

        # print(data)

        self.refresh_from_dict(
            job_dict = data['job'])
